use crate::compound::peano::{P0, Peano, PNext};

#[derive(Copy, Clone, Default, PartialEq)]
struct Null;

#[derive(Copy, Clone, Default, PartialEq)]
struct LLNode<E, N> {
    e: E,
    // Null or LLNode
    next: N,
}

impl<E> LLNode<E, Null> {
    fn new(e: E) -> Self {
        Self {
            e,
            next: Null,
        }
    }
}

impl<E, N> LLNode<E, N> {
    fn prepend<P>(self, e: P) -> LLNode<P, Self> {
        LLNode {
            e,
            next: self
        }
    }
}

struct ListBuilder<ElemTypes>(core::marker::PhantomData<ElemTypes>);
type List<ElemTypes> = <ListBuilder<ElemTypes> as ListBuilderOps>::Result;

trait ListBuilderOps {
    type Result;
}

impl<E0> ListBuilderOps for ListBuilder<(E0,)> {
    type Result = LLNode<E0, Null>;
}
impl<E0, E1> ListBuilderOps for ListBuilder<(E0, E1)> {
    type Result = LLNode<E0, List<(E1,)>>;
}
impl<E0, E1, E2> ListBuilderOps for ListBuilder<(E0, E1, E2)> {
    type Result = LLNode<E0, List<(E1, E2)>>;
}
impl<E0, E1, E2, E3> ListBuilderOps for ListBuilder<(E0, E1, E2, E3)> {
    type Result = LLNode<E0, List<(E1, E2, E3)>>;
}
impl<E0, E1, E2, E3, E4> ListBuilderOps for ListBuilder<(E0, E1, E2, E3, E4)> {
    type Result = LLNode<E0, List<(E1, E2, E3, E4)>>;
}
impl<E0, E1, E2, E3, E4, E5> ListBuilderOps for ListBuilder<(E0, E1, E2, E3, E4, E5)> {
    type Result = LLNode<E0, List<(E1, E2, E3, E4, E5)>>;
}
impl<E0, E1, E2, E3, E4, E5, E6> ListBuilderOps for ListBuilder<(E0, E1, E2, E3, E4, E5, E6)> {
    type Result = LLNode<E0, List<(E1, E2, E3, E4, E5, E6)>>;
}
impl<E0, E1, E2, E3, E4, E5, E6, E7> ListBuilderOps for ListBuilder<(E0, E1, E2, E3, E4, E5, E6, E7)> {
    type Result = LLNode<E0, List<(E1, E2, E3, E4, E5, E6, E7)>>;
}

/// implemented on a list to allow applying some function `F` to any (or all) elements of the list.
trait ListOp<F, List, Output> {
    /// apply `f` to all elements in the list.
    fn apply_all(&self, f: F);
    /// apply `f` *only* to the element at the provided index
    fn apply_at(&self, i: u32, f: F) -> Output {
        self.apply_at_rel(i, 0, f)
    }
    // helper: tracks the current index into the list to know when to call `f`
    fn apply_at_rel(&self, i: u32, cur: u32, f: F) -> Output;
}

/// some operation which can be applied to the provided element type `T`.
trait ElementOp<T> {
    type Output;
    /// perform the operation on the provided element.
    fn call(&self, el: &T) -> Self::Output;
}

impl<T, R, F: Fn(&T) -> R> ElementOp<T> for F {
    type Output = R;
    fn call(&self, el: &T) -> Self::Output {
        (self)(el)
    }
}

impl<Output, F, E, N> ListOp<F, LLNode<E, N>, Output> for LLNode<E, N>
where
    F: ElementOp<E, Output=Output>,
    N: ListOp<F, N, Output>
{
    fn apply_all(&self, f: F) {
        // apply the operation to our element
        f.call(&self.e);
        // apply the operation to all tail elements
        self.next.apply_all(f);
    }
    fn apply_at_rel(&self, i: u32, cur: u32, f: F) -> Output {
        if i == cur {
            f.call(&self.e)
        } else {
            self.next.apply_at_rel(i, cur + 1, f)
        }
    }
}

impl<Output, F> ListOp<F, Null, Output> for Null {
    fn apply_all(&self, _f: F) {
        // the tail element simply terminates recursion
    }
    fn apply_at_rel(&self, _i: u32, _cur: u32, _f: F) -> Output {
        // it's expected that the element was found by now
        unreachable!();
    }
}

/// implemented on a list AND index, allowing potentially non-generic code to fetch a
/// compile-time-constant element.
trait IndexOp<List, Index, Elem> {
    fn get(&self, idx: Index) -> &Elem;
}

/// zero-index into Self
impl<E, N> IndexOp<LLNode<E, N>, P0, E> for LLNode<E, N>
{
    fn get(&self, _idx: P0) -> &E {
        &self.e
    }
}

/// index N+1 into Self given that we know how to index N into Self::next
impl<E, N, P, ElemAtP> IndexOp<LLNode<E, N>, PNext<P>, ElemAtP> for LLNode<E, N>
where
    P: Peano,
    N: IndexOp<N, P, ElemAtP>
{
    fn get(&self, _idx: PNext<P>) -> &ElemAtP {
        self.next.get(P::default())
    }
}


#[cfg(test)]
mod test {
    use super::*;
    use crate::compound::peano::{P1, P2};
    use core::cell::Cell;

    #[derive(Default)]
    struct ApplyAccumulator<T> {
        delayed0: Cell<T>,
        delayed1: Cell<T>,
        delayed2: Cell<T>,
    }

    impl ElementOp<u32> for &ApplyAccumulator<u32> {
        type Output = ();
        fn call(&self, el: &u32) {
            self.delayed2.set(
                self.delayed1.replace(
                    self.delayed0.replace(*el)
                )
            )
        }

    }

    #[test]
    fn test_apply_single() {
        let list = LLNode::new(5u32);
        let acc = ApplyAccumulator::default();
        list.apply_all(&acc);
        
        assert_eq!(acc.delayed1.into_inner(), 0);
        assert_eq!(acc.delayed0.into_inner(), 5);
    }

    #[derive(Copy, Clone, Default, Debug, PartialEq)]
    struct V0(u32);
    #[derive(Copy, Clone, Default, Debug, PartialEq)]
    struct V1(f32);
    #[derive(Copy, Clone, Default, Debug, PartialEq)]
    struct Flat {
        v0: V0,
        v1: V1,
    }

    trait Variant {
        fn to_flat(&self) -> Flat;
    }
    impl Variant for V0 {
        fn to_flat(&self) -> Flat {
            Flat { v0: *self, ..Default::default() }
        }
    }
    impl Variant for V1 {
        fn to_flat(&self) -> Flat {
            Flat { v1: *self, ..Default::default() }
        }
    }

    impl<T: Variant> ElementOp<T> for &ApplyAccumulator<Flat> {
        type Output = ();
        fn call(&self, el: &T) {
            self.delayed2.set(
                self.delayed1.replace(
                    self.delayed0.replace(el.to_flat())
                )
            )
        }
    }

    #[test]
    fn test_apply_2() {
        let list = LLNode::new(V0(5u32))
            .prepend(V1(4f32));
        let acc = ApplyAccumulator::default();
        list.apply_all(&acc);
        
        assert_eq!(acc.delayed2.into_inner(), Flat::default());
        assert_eq!(acc.delayed1.into_inner(), V1(4f32).to_flat());
        assert_eq!(acc.delayed0.into_inner(), V0(5u32).to_flat());
    }

    #[test]
    fn test_apply_3() {
        let list = LLNode::new(V0(5u32))
            .prepend(V1(4f32))
            .prepend(V0(2u32));
        let acc = ApplyAccumulator::default();
        list.apply_all(&acc);
        
        assert_eq!(acc.delayed2.into_inner(), V0(2u32).to_flat());
        assert_eq!(acc.delayed1.into_inner(), V1(4f32).to_flat());
        assert_eq!(acc.delayed0.into_inner(), V0(5u32).to_flat());
    }

    #[test]
    fn test_apply_at() {
        let list = LLNode::new(V0(5u32))
            .prepend(V1(4f32))
            .prepend(V0(2u32));

        let acc = ApplyAccumulator::default();
        list.apply_at(0, &acc);
        assert_eq!(acc.delayed0.into_inner(), V0(2u32).to_flat());
        assert_eq!(acc.delayed1.into_inner(), Flat::default());

        let acc = ApplyAccumulator::default();
        list.apply_at(1, &acc);
        assert_eq!(acc.delayed0.into_inner(), V1(4f32).to_flat());
        assert_eq!(acc.delayed1.into_inner(), Flat::default());

        let acc = ApplyAccumulator::default();
        list.apply_at(2, &acc);
        assert_eq!(acc.delayed0.into_inner(), V0(5u32).to_flat());
        assert_eq!(acc.delayed1.into_inner(), Flat::default());
    }

    #[test]
    fn test_apply_at_ret() {
        struct ApplyAtRetOp;
        impl<T: Variant> ElementOp<T> for ApplyAtRetOp {
            type Output = Flat;
            fn call(&self, el: &T) -> Self::Output {
                el.to_flat()
            }
        }

        let list = LLNode::new(V0(5u32))
            .prepend(V1(4f32))
            .prepend(V0(2u32));

        assert_eq!(list.apply_at(0, ApplyAtRetOp), V0(2u32).to_flat());
        assert_eq!(list.apply_at(1, ApplyAtRetOp), V1(4f32).to_flat());
        assert_eq!(list.apply_at(2, ApplyAtRetOp), V0(5u32).to_flat());
    }

    #[test]
    fn get() {
        let list = LLNode::new(3f32)
            .prepend(4i32)
            .prepend(5u32);
        assert_eq!(list.get(P0), &5u32);
        assert_eq!(list.get(P1::default()), &4i32);
        assert_eq!(list.get(P2::default()), &3f32);
    }

    #[test]
    fn test_list_builder() {
        let type_level = List::<(i32, f32, u32)>::default();
        let value_level = LLNode::new(0u32)
            .prepend(0f32)
            .prepend(0i32);

        assert!(type_level == value_level);
    }
}
