use crate::mat::{DiscriminantCodable, Material};
use crate::real::Real;
use crate::vec::Vec3;

#[cfg(feature = "serde")]
use serde::{Serialize, Deserialize};

#[cfg_attr(feature = "serde", derive(Deserialize, Serialize))]
#[cfg_attr(feature = "fmt", derive(Debug))]
#[derive(Copy, Clone, Default, PartialEq)]
pub struct Conductor<T>(T);
pub type AnisomorphicConductor<R> = Conductor<Vec3<R>>;
pub type IsomorphicConductor<R> = Conductor<(R,)>;

impl<R> IsomorphicConductor<R> {
    pub fn new(c: R) -> Self {
        Self((c,))
    }
}

impl<V: Clone> IsomorphicConductor<V> {
    pub fn iso_conductivity(&self) -> V {
        self.0.0.clone()
    }
}

impl<R> AnisomorphicConductor<R> {
    pub fn new(c: Vec3<R>) -> Self {
        Self(c)
    }
}

impl<R: Real> Into<AnisomorphicConductor<R>> for IsomorphicConductor<R> {
    fn into(self) -> AnisomorphicConductor<R> {
        AnisomorphicConductor::new(Vec3::uniform(self.iso_conductivity()))
    }
}

impl<R: Real> Material<R> for AnisomorphicConductor<R> {
    fn conductivity(&self) -> Vec3<R> {
        self.0
    }
}

impl<R: Real> Material<R> for IsomorphicConductor<R> {
    fn conductivity(&self) -> Vec3<R> {
        Vec3::uniform(self.iso_conductivity())
    }
}

impl<R: Real> DiscriminantCodable for IsomorphicConductor<R> {
    fn discr(&self) -> u32 {
        let cond = self.iso_conductivity();
        if cond < R::zero() {
            (-cond.to_f32()) as u32
        } else {
            0
        }
    }
    fn make_discr(d: u32) -> Self {
        Self::new(R::from_primitive(-(d as i32)))
    }
}
