use crate::real::Real;
use crate::vec::Vec3;

mod conductor;
mod mb_pgram;
mod mh_pgram;
pub use conductor::{AnisomorphicConductor, IsomorphicConductor};
pub use mb_pgram::MBPgram;
pub use mh_pgram::{Ferroxcube3R1MH, MHPgram};

#[cfg(feature = "serde")]
use serde::{Serialize, Deserialize};


pub trait Material<R: Real>: Sized {
    fn conductivity(&self) -> Vec3<R> {
        Default::default()
    }
    /// returns the new M vector for this material
    fn move_b_vec(&self, m: Vec3<R>, _target_b: Vec3<R>) -> Vec3<R> {
        // XXX could return either 0, or `m`. they should be the same, but one might be more
        // optimizable than the other (untested).
        m
    }
}

/// any type which may hold a discriminant internally.
pub trait DiscriminantCodable {
    fn discr(&self) -> u32;
    fn make_discr(d: u32) -> Self;
}

pub struct D0<T>(pub T);
pub struct D1<T>(pub T);

#[cfg_attr(feature = "serde", derive(Deserialize, Serialize))]
#[cfg_attr(feature = "fmt", derive(Debug))]
#[derive(Copy, Clone, Default, PartialEq)]
pub struct DiscrMat2<M0, M1> {
    m0: M0,
    m1: M1,
}

impl<M0, M1> DiscrMat2<M0, M1> {
    fn new_from_fields(m0: M0, m1: M1) -> Self {
        Self { m0, m1 }
    }
}

impl<M0: DiscriminantCodable, M1: Default> DiscrMat2<M0, M1> {
    pub fn new0(m: M0) -> Self {
        D0(m).into()
    }
    pub fn new1(m: M1) -> Self {
        D1(m).into()
    }
}

impl<M0: DiscriminantCodable, M1: Default> From<D0<M0>> for DiscrMat2<M0, M1> {
    fn from(m0: D0<M0>) -> Self {
        let m0 = m0.0;
        assert!(m0.discr() == 0, "unexpected non-zero discriminant");
        Self::new_from_fields(m0, Default::default())
    }
}

impl<M0: DiscriminantCodable, M1> From<D1<M1>> for DiscrMat2<M0, M1> {
    fn from(m1: D1<M1>) -> Self {
        let m1 = m1.0;
        Self::new_from_fields(M0::make_discr(1), m1)
    }
}

impl<R: Real, M0: DiscriminantCodable + Material<R>, M1: Material<R> + Copy> Material<R> for DiscrMat2<M0, M1> {
    fn conductivity(&self) -> Vec3<R> {
        match self.m0.discr() {
            0 => self.m0.conductivity(),
            1 => {
                let mat = self.m1;  //< XXX hack for ZST
                mat.conductivity()
            }
            _ => panic!()
            // _ => unreachable!(),
        }
    }
    fn move_b_vec(&self, m: Vec3<R>, target_b: Vec3<R>) -> Vec3<R> {
        match self.m0.discr() {
            0 => self.m0.move_b_vec(m, target_b),
            1 => {
                let mat = self.m1;  //< XXX hack for ZST
                mat.move_b_vec(m, target_b)
            }
            _ => panic!()
            // _ => unreachable!(),
        }
    }
}

pub type IsoConductorOr<R, M1> = DiscrMat2<IsomorphicConductor<R>, M1>;

impl<R: Real, M1: Default> IsoConductorOr<R, M1> {
    pub fn new_conductor(conductivity: R) -> Self {
        Self::new0(IsomorphicConductor::new(conductivity))
    }
    pub fn new_other(other: M1) -> Self {
        Self::new1(other)
    }
}

// XXX: can't do this for generic M, because that creates duplicate `From` impls for the
// IsomorphicConductor itself
impl<R: Real> From<Ferroxcube3R1MH> for IsoConductorOr<R, Ferroxcube3R1MH> {
    fn from(mat: Ferroxcube3R1MH) -> Self {
        IsoConductorOr::new1(mat)
    }
}
impl<R: Real> From<IsomorphicConductor<R>> for IsoConductorOr<R, Ferroxcube3R1MH> {
    fn from(mat: IsomorphicConductor<R>) -> Self {
        IsoConductorOr::new0(mat)
    }
}

#[cfg(test)]
mod test {
    use super::*;
    use crate::real::R32;

    #[test]
    fn iso_conductor_discr() {
        type T = IsomorphicConductor<f32>;
        let c = T::make_discr(5);
        assert_eq!(c.discr(), 5);

        let c = T::make_discr(0);
        assert_eq!(c.discr(), 0);

        let c = T::new(5.0);
        assert_eq!(c.discr(), 0);

        let c = T::new(0.0);
        assert_eq!(c.discr(), 0);
    }

    #[test]
    fn iso_conductor_or_3r1() {
        let c: IsoConductorOr<f32, Ferroxcube3R1MH> = IsomorphicConductor::new(22.0f32).into();
        assert!(c.conductivity() == Vec3::uniform(22.0f32));

        let c: IsoConductorOr<f32, Ferroxcube3R1MH> = Ferroxcube3R1MH::new().into();
        assert!(c.conductivity() == Vec3::zero());
    }

    #[test]
    fn iso_conductor_or_aniso() {
        type I = IsoConductorOr<R32, AnisomorphicConductor<R32>>;

        let c = I::new0(IsomorphicConductor::new(22f32.cast()));
        assert!(c.conductivity() == Vec3::uniform(22.0).cast());

        let c = I::new1(AnisomorphicConductor::new(
            Vec3::new(2.0, 3.0, 4.0).cast()
        ));

        assert!(c.conductivity() == Vec3::new(2.0, 3.0, 4.0).cast());
    }
}
